import json
from pathlib import Path

import pandas as pd

from . import constants as const


def dump_readme():
    lpath = Path(const.LOG_DIR)
    df = pd.DataFrame(
        map(json.loads, map(Path.read_text, lpath.glob("*.json")))
    )
    fresh_things = (
        df.sort_values("at")
        .groupby(["commit", "name", "input_id"])
        .last()
        .reset_index()
        .pivot_table(index=["name", "commit"], columns="input_id")
    )
    succ_ser = fresh_things["is_success"].fillna(0).all(axis=1)
    good_ones = (
        fresh_things.loc[succ_ser, "duration"]
        .loc[:, lambda df: df.mean().sort_values().index]
        .assign(**{"Total time": lambda df: df.sum(axis=1)})
        .sort_values("Total time")
        .round(4)
    )

    top_html = good_ones.to_html()
    bot_html = fresh_things.loc[~succ_ser, :].to_html()

    out_str = "\n\n".join(
        [
            "# Results",
            "## Successful Solutions",
            top_html,
            "## Near Misses",
            bot_html,
        ]
    )
    Path("README.md").write_text(out_str)



import sys

from .data_challenge.setup_dir import create_dir


def main(args):
    try:
        arg = args[1]
    except IndexError:
        arg = ""
    if arg == "challenge":
        try:
            target = args[3]
        except IndexError:
            target = "."
        create_dir(args[2], target)


if __name__ == "__main__":  # pragma: no cover
    main(sys.argv)



import os
import sys
from contextlib import contextmanager

from invoke import UnexpectedExit


def git_commit(c, addstr, msg):
    try:
        c.run("git config --get user.email")
        c.run("git config --get user.name")
    except UnexpectedExit:
        c.run('git config --local user.email "ci@cd.org"')
        c.run('git config --local user.name "CI/CD"')
    c.run(f'git add {addstr} && git commit -m "{msg}"')


@contextmanager
def cd_into(dirpath):
    wd = os.getcwd()
    os.chdir(dirpath)
    sys.path.insert(0, str(dirpath))
    yield
    os.chdir(wd)
    sys.path.pop(0)



from pathlib import Path

import pytest
from invoke import Context

from erudition.__main__ import main
from erudition.data_challenge import constants, tasks
from erudition.util import cd_into, git_commit


def test_full(tmp_path):
    c = Context()
    origin_path = tmp_path / "origin"
    origin_path.mkdir(exist_ok=True)
    with cd_into(origin_path):
        c.run("git init -b main")
        c.run("git checkout -b tmp")

    ch_name = "test-challenge"
    main([None, "challenge", ch_name, tmp_path])

    with cd_into(tmp_path / f"{ch_name}-challenge"):
        c.run(f"git remote add origin {origin_path}")
        c.run("git push -u origin main")
        with cd_into(origin_path):
            c.run("git checkout main")
            with pytest.raises(EnvironmentError):
                tasks.test_modified_solutions(c, "P-1")
            c.run("git checkout -b tmp")
        c.run("git push --tags origin")
        _dummy_solution("s1", "{1: 2}")
        git_commit(c, "s1", "add s1")
        tasks.test_modified_solutions(c, "P-1")
        tasks.retag(c)
        _dummy_solution("s2", '{"X": 0}')
        git_commit(c, "s2", "add s2")
        tasks.test_modified_solutions(c, "P-1")
        tasks.retag(c)
        tasks.test_modified_solutions(c, "P-1")
        _dummy_solution("s2", '{"A": 10}')
        tasks.test_solution(c, "s2", "P-1")
        tasks.get_test_pack(c)


def _dummy_solution(name, todump):
    Path(name).mkdir(exist_ok=True)
    Path(name, constants.CONF_PATH).write_text("process-command: python m.py")
    py_str = f"""open("output.json", "w").write('{todump}')"""
    Path(name, "m.py").write_text(py_str)



CHALLENGE_YAML = "challenge.yaml"
PRIV_PACKID_KEY = "private-ids"
PUB_PACKID_KEY = "public-ids"
LOG_DIR = ".logs"

CONF_PATH = "conf.yaml"
SETUP_COMM = "setup-command"
ETL_COMM = "etl-command"
PROC_COMM = "process-command"


RESULTS_FILENAME = "results.json"
INPUT_FILENAME = "input.json"

SPEC_MODULE = "chspec"
PACK_FUNCTION = "load_pack"
EVAL_FUNCTION = "evaluate"


PR_GHA_PATH, PUSH_GHA_PATH = [
    f".github/workflows/run-{k}.yml" for k in ["pr", "full"]
]

EVALED_GIT_TAG = "eval-checkpoint"



import io
import json
from contextlib import redirect_stdout
from datetime import datetime
from distutils.dir_util import copy_tree
from functools import partial, reduce
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from uuid import uuid4
from zipfile import ZipFile

from invoke import UnexpectedExit, task
from structlog import get_logger

from ..util import git_commit
from . import constants as const
from .draw import dump_readme
from .util import CConf, get_obj

_RUNNER = "__r.sh"
_TIMER = "__t"

logger = get_logger(ctx="data challenge task")


class PackRepo:
    def __init__(self, pack_id) -> None:

        pack_loader = get_obj(const.PACK_FUNCTION)
        self.tmpfile = NamedTemporaryFile()
        pack_loader(pack_id, self.tmpfile.name)

    def dump_data(self, dirname):
        self._dump(dirname, [const.INPUT_FILENAME, const.RESULTS_FILENAME])

    def dump_inputs(self, dirname):
        self._dump(dirname, only=const.INPUT_FILENAME)

    def dump_results(self, dirname):
        self._dump(dirname, only=const.RESULTS_FILENAME)

    def cleanup(self):
        self.tmpfile.close()

    def _dump(self, dirname, exclude=(), only=None):
        with ZipFile(self.tmpfile.name) as zip_path:
            for cfp in zip_path.filelist:
                _name = cfp.filename
                if (_name in exclude) or (only and (_name != only)):
                    continue
                zip_path.extract(cfp, dirname)


@task
def test_solution(c, name, input_id, fail=True):
    pack_repo = PackRepo(input_id)
    _eval(c, name, pack_repo, input_id, fail, False)
    pack_repo.cleanup()


@task
def test_modified_solutions(c, input_id, fail=False, push_logs=True):
    changed_solutions = _get_changes(c)
    if not changed_solutions:
        return
    pack_repo = PackRepo(input_id)
    for solution in changed_solutions:
        _eval(c, solution, pack_repo, input_id, fail, push_logs)
    pack_repo.cleanup()


@task
def get_test_pack(c):
    get_obj(const.PACK_FUNCTION)()


@task
def retag(c):
    tag_name = f"{const.EVALED_GIT_TAG}-{uuid4().hex}"
    c.run(f"git tag {tag_name}")
    c.run(f"git push origin {tag_name}")
    try:
        dump_readme()
        git_commit(c, "README.md", "update readme")
    except Exception as e:
        logger.exception(e)


def _eval(c, solution_name, pack_repo: PackRepo, input_id, fail, push):
    logger.info(f"evaluating solution {solution_name} on input {input_id}")
    sdir = Path.cwd() / solution_name
    contid = "challenge_dcont_eu"
    conf = CConf.load(solution_name)
    tmpdir = TemporaryDirectory()
    dirname = tmpdir.name
    eval_fun = get_obj(const.EVAL_FUNCTION)

    copy_tree(sdir, dirname)
    c.run(f"docker run -v {dirname}:/work --name {contid} -dt {conf.image}")
    _run = partial(_runcmd, c, contid)
    try:
        _run(conf.setup)
        pack_repo.dump_data(dirname)
        _run(conf.etl)
        pack_repo.dump_inputs(dirname)
        proc_time = _timed(conf.process, _run, dirname)
        pack_repo.dump_results(dirname)
        succ = eval_fun(
            dirname,
            json.loads((Path(dirname) / const.RESULTS_FILENAME).read_text()),
        )
    except Exception as e:
        logger.exception(e)
        proc_time = float("inf")
        succ = False
    finally:
        c.run(f"docker kill {contid} && docker rm {contid}")
        tmpdir.cleanup()
        if fail:
            assert succ
        _log(c, solution_name, input_id, succ, proc_time, _gethash(c), push)


def _runcmd(c, containerid, comm):
    c.run(f"docker exec -w /work {containerid} {comm}")


def _timed(comm, runner, dirpath):
    timecomm = f"date +%s.%N >> {_TIMER}"
    sh_str = "\n".join(["#!/bin/sh", timecomm, comm, timecomm])
    Path(dirpath, _RUNNER).write_text(sh_str)
    runner(f"chmod +x {_RUNNER}")
    runner(f"./{_RUNNER}")
    start_time, end_time = map(
        float, Path(dirpath, _TIMER).read_text().strip().split("\n")
    )
    return end_time - start_time


def _gethash(c):
    f = io.StringIO()
    try:
        with redirect_stdout(f):
            c.run("git rev-parse --short HEAD")
        return f.getvalue().strip()
    except UnexpectedExit:
        return


def _log(c, solution_name, input_id, result, proc_time, commit_hash, push):
    logdic = {
        "name": solution_name,
        "input_id": input_id,
        "is_success": result,
        "duration": proc_time,
        "commit": commit_hash,
        "at": datetime.now().isoformat(),
    }
    logstr = json.dumps(logdic)
    logger.info("DONE", **logdic)
    lpath = Path(const.LOG_DIR)
    lpath.mkdir(exist_ok=True)
    log_id = uuid4().hex
    (lpath / f"{log_id}.json").write_text(logstr)
    git_commit(c, const.LOG_DIR, f"add logs {log_id[:8]}")
    if push:
        c.run("git config --local pull.rebase true")
        c.run("git pull; git push")


def _get_changes(c):
    tags = filter(
        lambda s: s.startswith(const.EVALED_GIT_TAG),
        _get_lines(c, "git tag"),
    )
    try:
        fun = partial(_get_diff_dirs, c)
        return [*reduce(set.intersection, map(fun, tags))]
    except TypeError:
        msg = f"no {const.EVALED_GIT_TAG} in {_get_lines(c, 'git tag')}"
        raise EnvironmentError(msg)


def _get_lines(c, comm):
    f = io.StringIO()
    with redirect_stdout(f):
        c.run(comm)
    return f.getvalue().strip().split("\n")


def _get_diff_dirs(c, base_commit):
    changes = set()
    for poss_ch in _get_lines(c, f"git diff {base_commit}..HEAD --name-only"):
        if (
            poss_ch.startswith(".")
            or poss_ch.startswith("__")
            or (not poss_ch)
        ):
            continue
        poss_dir = Path(poss_ch).parts[0]
        if Path(poss_dir).is_dir() and Path(poss_dir).exists():
            changes.add(poss_dir)
    return changes



from dataclasses import dataclass
from importlib import import_module
from pathlib import Path

from yaml import safe_load

from . import constants as const


@dataclass
class CConf:
    process: str
    setup: str = "sleep 0"
    etl: str = "sleep 0"
    image: str = "python:3.9"

    @classmethod
    def load(cls, solution):
        dic = safe_load((Path(solution) / const.CONF_PATH).read_text())
        return cls(**{k.split("-")[0]: v for k, v in dic.items() if v})


def get_obj(obj_name):
    return getattr(import_module(const.SPEC_MODULE), obj_name)



import json
from distutils.dir_util import copy_tree
from pathlib import Path
from tempfile import TemporaryDirectory

from invoke import Context
from yaml import safe_load

from ..util import cd_into, git_commit
from . import constants as const

root_repo = "https://github.com/endremborza/teaching"


def create_dir(challenge_name: str, target="."):
    c = Context()
    tmp_dir = TemporaryDirectory()
    c.run(f"git clone {root_repo} {tmp_dir.name}")
    ch_root = Path(tmp_dir.name) / "data-challenges"
    frame_dir = ch_root / "frame"
    ch_dir = ch_root / challenge_name
    out_dir = (Path(target) / f"{challenge_name}-challenge").as_posix()
    copy_tree(frame_dir.as_posix(), out_dir)
    copy_tree(ch_dir.as_posix(), out_dir)

    _conf = safe_load(Path(out_dir, const.CHALLENGE_YAML).read_text())
    priv_ids = _conf[const.PRIV_PACKID_KEY]
    pub_ids = _conf[const.PUB_PACKID_KEY]

    for _path, _pids in [
        (const.PR_GHA_PATH, pub_ids),
        (const.PUSH_GHA_PATH, pub_ids + priv_ids),
    ]:
        a_path = Path(out_dir, _path)
        a_str = a_path.read_text().replace("[...]", json.dumps(_pids))
        a_path.write_text(a_str)
    tmp_dir.cleanup()

    with cd_into(out_dir):
        c.run("git init -b main")
        git_commit(c, "* .github .gitignore", "init")
        c.run(f"git tag {const.EVALED_GIT_TAG}-v0")
